/***************************************************************************
 *   Copyright (C) 2007 by Dr. Feynman                                     *
 *   dr@feynman.com                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "feynman.h"

#ifdef WIN32
#include "feynman_win32.h"
#endif

int feynman_send_login(t_feynman_remote_server *server)
{
    unsigned char buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    if(!server) return FEYNMAN_ERROR;

    memset(buf, 0, sizeof(buf));

    buf[i++] = FEYNMAN_MSG_LOGIN;
    buf[i++] = 0;
    buf[i++] = 0;

    buf[i++] = FEYNMAN_DATA_USERNAME;
    buf[i++] = strlen(server->username);

    strcpy((char *)buf+i, server->username);
    i += strlen(server->username);

    buf[i++] = FEYNMAN_DATA_PASSWORD;
    buf[i++] = strlen(server->password);

    strcpy((char *)buf+i, server->password);
    i += strlen(server->password);

    buf[1] = ((i-3) >> 8);
    buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(server->handle, buf, i) < 0) return FEYNMAN_ERROR;
    return FEYNMAN_STATUS_OK;
}

int feynman_process_login_ack(t_feynman_remote_server *server, unsigned char *buf, int len)
{
    server->access_mode = buf[0];
    if(server->access_mode == FEYNMAN_ACCESS_MODE_NOACCESS) return FEYNMAN_STATUS_NOK;

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Access OK\n");

    if(!(server->access_mode & FEYNMAN_ACCESS_MODE_EMM))
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] EMM: No\n");
    }
    else
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] EMM: Yes\n");
    }
    server->logged_in = 1;
    return FEYNMAN_STATUS_OK;
}

int feynman_send_login_ack(t_feynman_client *client, t_feynman_handle *handle)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    send_buf[i++] = FEYNMAN_MSG_LOGIN_ACK;
    send_buf[i++] = 0;
    send_buf[i++] = 0;

    send_buf[i] = FEYNMAN_ACCESS_MODE_NOACCESS;

    if(client != NULL)
    {
        send_buf[i] = client->access_mode;
        client->handle.fd = handle->fd;
#ifdef HAVE_SSL
        client->handle.ssl = handle->ssl;
#endif
    }
    i++;

    send_buf[1] = ((i-3) >> 8);
    send_buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(*handle, send_buf, i) < 0) return FEYNMAN_ERROR;
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Login ack sent\n");

    return FEYNMAN_STATUS_OK;
}

t_feynman_client *feynman_process_login(t_feynman_local_server *server, unsigned char *buf, int len)
{
    t_feynman_client *client;
    int i;
    char username[255];
    char password[255];

    if(!server) return NULL;

    memset(username, 0, sizeof(username));
    memset(password, 0, sizeof(password));

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];

        switch(nano)
        {
            case FEYNMAN_DATA_USERNAME:
            {
                memcpy(username, buf+i, nanolen);
                username[nanolen] = '\0';
                break; 
            }
            case FEYNMAN_DATA_PASSWORD:
            {
                memcpy(password, buf+i, nanolen);
                password[nanolen] = '\0';
                break; 
            }
        }
        i += nanolen;
    }

    for(client=server->client_db; client!=NULL; client=client->next)
    {
        if(!strcmp(client->username, username))
        {
            if(!strcmp(client->password, password))
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] %s logged in\n", username);
                return client;
            }
        }
    }

    return NULL;
}

int feynman_send_add_card(t_feynman_client *client, t_feynman_card *card)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    if(card == NULL) return FEYNMAN_ERROR;
    if(client == NULL) return FEYNMAN_ERROR;

    if(!(client->access_mode & FEYNMAN_ACCESS_MODE_ACCESSOK)) return FEYNMAN_STATUS_NOK;

    send_buf[i++] = FEYNMAN_MSG_ADD_CARD;
    send_buf[i++] = 0;
    send_buf[i++] = 0;

    send_buf[i++] = FEYNMAN_DATA_CARD_ID;
    send_buf[i++] = 4;
    send_buf[i++] = (card->card_id >> 24) & 0xFF;
    send_buf[i++] = (card->card_id >> 16) & 0xFF;
    send_buf[i++] = (card->card_id >> 8) & 0xFF;
    send_buf[i++] = card->card_id & 0xFF;

    send_buf[i++] = FEYNMAN_DATA_CA_ID;
    send_buf[i++] = 2;
    send_buf[i++] = (card->caid >> 8) & 0xFF;
    send_buf[i++] = card->caid & 0xFF;

    send_buf[i++] = FEYNMAN_DATA_PROVIDER_ID;
    send_buf[i++] = 4;
    send_buf[i++] = (card->provid >> 24) & 0xFF;
    send_buf[i++] = (card->provid >> 16) & 0xFF;
    send_buf[i++] = (card->provid >> 8) & 0xFF;
    send_buf[i++] = card->provid & 0xFF;

    if(client->access_mode & FEYNMAN_ACCESS_MODE_GA)
    {
        send_buf[i++] = FEYNMAN_DATA_GA;
        send_buf[i++] = 4;
        send_buf[i++] = (card->ga >> 24) & 0xFF;
        send_buf[i++] = (card->ga >> 16) & 0xFF;
        send_buf[i++] = (card->ga >> 8) & 0xFF;
        send_buf[i++] = card->ga & 0xFF;
    }

    if(client->access_mode & FEYNMAN_ACCESS_MODE_SA)
    {
        send_buf[i++] = FEYNMAN_DATA_SA;
        send_buf[i++] = 4;
        send_buf[i++] = (card->sa >> 24) & 0xFF;
        send_buf[i++] = (card->sa >> 16) & 0xFF;
        send_buf[i++] = (card->sa >> 8) & 0xFF;
        send_buf[i++] = card->sa & 0xFF;
    }

    if(client->access_mode & FEYNMAN_ACCESS_MODE_UA)
    {
        send_buf[i++] = FEYNMAN_DATA_UA;
        send_buf[i++] = 4;
        send_buf[i++] = (card->ua >> 24) & 0xFF;
        send_buf[i++] = (card->ua >> 16) & 0xFF;
        send_buf[i++] = (card->ua >> 8) & 0xFF;
        send_buf[i++] = card->ua & 0xFF;
    }

    send_buf[1] = ((i-3) >> 8);
    send_buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(client->handle, send_buf, i) < 0) return FEYNMAN_ERROR;
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Card id %X sent to %s\n", card->card_id, client->username);
    return FEYNMAN_STATUS_OK;
}

int feynman_process_add_card(t_feynman_remote_server *server, unsigned char *buf, int len)
{
    int i;
    t_feynman_card *card = NULL;

    if(!server) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];
        switch(nano)
        {
            case FEYNMAN_DATA_CARD_ID:
            {
                int card_id = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                card = feynman_add_remote_card(server, card_id);
                break;
            }
            case FEYNMAN_DATA_CA_ID:
            {
                if(card == NULL) return FEYNMAN_ERROR;
                unsigned short caid = (buf[i] << 8) | (buf[i+1] << 0);
                card->caid = caid;
                break;
            }
            case FEYNMAN_DATA_PROVIDER_ID:
            {
                if(card == NULL) return FEYNMAN_ERROR;
                int provid = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                card->provid = provid;
                break;
            }
            case FEYNMAN_DATA_GA:
            {
                if(card == NULL) return FEYNMAN_ERROR;
                int ga = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                card->ga = ga;
                break;
            }
            case FEYNMAN_DATA_SA:
            {
                if(card == NULL) return FEYNMAN_ERROR;
                int sa = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                card->sa = sa;
                break;
            }
            case FEYNMAN_DATA_UA:
            {
                if(card == NULL) return FEYNMAN_ERROR;
                int ua = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                card->ua = ua;
                break;
            }
        }
        i += nanolen;
    }
    if(card != NULL)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Added card %X: caid: %04X provid:%08X ga: %08X sa: %08X ua: %08X\n", card->card_id, card->caid, card->provid, card->ga, card->sa, card->ua);
        return FEYNMAN_STATUS_OK;
    }
    return FEYNMAN_ERROR;
}

int feynman_send_del_card(t_feynman_client *client, t_feynman_card *card)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    if(card == NULL) return FEYNMAN_ERROR;
    if(client == NULL) return FEYNMAN_ERROR;

    if(!(client->access_mode & FEYNMAN_ACCESS_MODE_ACCESSOK)) return FEYNMAN_STATUS_NOK;

    send_buf[i++] = FEYNMAN_MSG_DEL_CARD;
    send_buf[i++] = 0;
    send_buf[i++] = 0;

    send_buf[i++] = FEYNMAN_DATA_CARD_ID;
    send_buf[i++] = 4;
    send_buf[i++] = (card->card_id >> 24) & 0xFF;
    send_buf[i++] = (card->card_id >> 16) & 0xFF;
    send_buf[i++] = (card->card_id >> 8) & 0xFF;
    send_buf[i++] = card->card_id & 0xFF;

    send_buf[1] = ((i-3) >> 8);
    send_buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(client->handle, send_buf, i) < 0) return FEYNMAN_ERROR;
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Told %s to delete card %X\n", client->username, card->card_id);
    return FEYNMAN_STATUS_OK;
}


int feynman_process_del_card(t_feynman_remote_server *server, unsigned char *buf, int len)
{
    int i;

    if(!server) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];

        switch(nano)
        {
            case FEYNMAN_DATA_CARD_ID:
            {
                int card_id = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);

                feynman_del_remote_card(server, card_id);
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Deleted card %X\n", card_id);
                break; 
            }
        }
        i += nanolen;
    }

    return FEYNMAN_STATUS_OK;
}


int feynman_send_sid(t_feynman_client *client, int card_id, t_feynman_sid *sid)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    if(client == NULL) return FEYNMAN_ERROR;
    if(sid == NULL) return FEYNMAN_ERROR;
    if(card_id < 0) return FEYNMAN_ERROR;

    if(!(client->access_mode & FEYNMAN_ACCESS_MODE_ACCESSOK)) return FEYNMAN_STATUS_NOK;

    send_buf[i++] = FEYNMAN_MSG_SIDS;
    send_buf[i++] = 0;
    send_buf[i++] = 0;

    send_buf[i++] = FEYNMAN_DATA_CARD_ID;
    send_buf[i++] = 4;
    send_buf[i++] = (card_id >> 24) & 0xFF;
    send_buf[i++] = (card_id >> 16) & 0xFF;
    send_buf[i++] = (card_id >> 8) & 0xFF;
    send_buf[i++] = card_id & 0xFF;

    send_buf[i++] = FEYNMAN_DATA_SID;
    send_buf[i++] = 3;
    send_buf[i++] = sid->mode;
    send_buf[i++] = (sid->sid >> 8) & 0xFF;
    send_buf[i++] = sid->sid & 0xFF;

    send_buf[1] = ((i-3) >> 8);
    send_buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(client->handle, send_buf, i) < 0) return FEYNMAN_ERROR;

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Sid %04X for card %X sent to %s\n", sid->sid, card_id, client->username);
    return FEYNMAN_STATUS_OK;
}

int feynman_send_sidlist(t_feynman_client *client, t_feynman_card *card)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];

    if(client == NULL) return FEYNMAN_ERROR;
    if(card == NULL) return FEYNMAN_ERROR;

    if(!(client->access_mode & FEYNMAN_ACCESS_MODE_ACCESSOK)) return FEYNMAN_STATUS_NOK;

    t_feynman_sid *s = card->sid_db;
    if(s == NULL) return FEYNMAN_STATUS_NOK;

    do
    {
        int i = 0;

        send_buf[i++] = FEYNMAN_MSG_SIDS;
        send_buf[i++] = 0;
        send_buf[i++] = 0;

        send_buf[i++] = FEYNMAN_DATA_CARD_ID;
        send_buf[i++] = 4;
        send_buf[i++] = (card->card_id >> 24) & 0xFF;
        send_buf[i++] = (card->card_id >> 16) & 0xFF;
        send_buf[i++] = (card->card_id >> 8) & 0xFF;
        send_buf[i++] = card->card_id & 0xFF;

        send_buf[i++] = FEYNMAN_DATA_SID;
        int sid_len_pos = i;
        send_buf[i++] = 0;

        while(s != NULL && i < 250)
        {
            send_buf[i++] = s->mode;
            send_buf[i++] = (s->sid >> 8) & 0xFF;
            send_buf[i++] = s->sid & 0xFF;

            send_buf[sid_len_pos] += 3;
            s = s->next;
        }

        send_buf[1] = ((i-3) >> 8);
        send_buf[2] = ((i-3) >> 0);
        if(feynman_send_network_message(client->handle, send_buf, i) < 0) return FEYNMAN_ERROR;

    } while(s != NULL);

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Sidlist for card %X sent to %s\n", card->card_id, client->username);
    return FEYNMAN_STATUS_OK;
}

int feynman_process_sids(t_feynman_remote_server *server, unsigned char *buf, int len)
{
    int i;
    t_feynman_card *card = NULL;

    if(server == NULL) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];

        switch(nano)
        {
            case FEYNMAN_DATA_CARD_ID:
            {
                int card_id = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                card = feynman_get_remote_card(server, card_id);
                break; 
            }
            case FEYNMAN_DATA_SID:
            {
                if(card == NULL) return FEYNMAN_ERROR;
                int j;
                for(j=0; j<nanolen; j+=3)
                { 
                    char sidmode = buf[i+j];
                    unsigned short new_sid = (buf[i+j+1] << 8) | (buf[i+j+2] << 0);
                    t_feynman_sid *sid = feynman_get_sid(card, new_sid);
                    if(sid)
                    { 
                        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Changed mode on sid %04X, card %X (%d --> %d)\n", new_sid, card->card_id, sid->mode, sidmode);
                        sid->mode = sidmode;
                    }
                    else
                    { 
                        feynman_add_sid(card, new_sid, sidmode);
                        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Added sid %04X on card %X\n", new_sid, card->card_id);
                    }
                }
                break; 
            }
        }
        i += nanolen;
    }
    if(card != NULL)
    {
        return FEYNMAN_STATUS_OK;
    }

    return FEYNMAN_ERROR;
}

int feynman_send_ecm(t_feynman_remote_server *server, t_feynman_ecm *ecm)
{
    unsigned char buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    if(server == NULL) return FEYNMAN_ERROR;
    if(ecm == NULL) return FEYNMAN_ERROR;

    if(!(server->access_mode & FEYNMAN_ACCESS_MODE_ECM))
    {
        return FEYNMAN_STATUS_NOK;
    }

    buf[i++] = FEYNMAN_MSG_ECM;
    buf[i++] = 0;
    buf[i++] = 0;

    buf[i++] = FEYNMAN_DATA_CARD_ID;
    buf[i++] = 4;
    buf[i++] = (ecm->card_id >> 24) & 0xFF;
    buf[i++] = (ecm->card_id >> 16) & 0xFF;
    buf[i++] = (ecm->card_id >> 8) & 0xFF;
    buf[i++] = ecm->card_id & 0xFF;

    buf[i++] = FEYNMAN_DATA_SESSION_ID;
    buf[i++] = 2;
    buf[i++] = (ecm->session >> 8) & 0xFF;
    buf[i++] = ecm->session & 0xFF;

    buf[i++] = FEYNMAN_DATA_SID;
    buf[i++] = 2;
    buf[i++] = (ecm->sid >> 8) & 0xFF;
    buf[i++] = ecm->sid & 0xFF;

    buf[i++] = FEYNMAN_DATA_ECM;
    buf[i++] = ecm->ecm_len;
    memcpy(buf+i, ecm->ecm, ecm->ecm_len);
    i += ecm->ecm_len;

    buf[1] = ((i-3) >> 8);
    buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(server->handle, buf, i) < 0) return FEYNMAN_ERROR;
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] ECM session %d sent to card %X\n", ecm->session, ecm->card_id);
    return FEYNMAN_STATUS_OK;
}

int feynman_process_ecm(t_feynman_client *client, t_feynman_local_server *server, unsigned char *buf, int len)
{
    unsigned char ecm_data[FEYNMAN_BUF_SIZE];

    int sid = -1;
    int card_id = -1;
    int ecm_len = -1;
    int session_id = -1;
    int i;

    if(client == NULL) return FEYNMAN_ERROR;
    if(server == NULL) return FEYNMAN_ERROR;

    if(!(client->access_mode & FEYNMAN_ACCESS_MODE_ECM))
    {
        return FEYNMAN_STATUS_NOK;
    }

    if(server->ecm_callback == NULL) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];

        switch(nano)
        {
            case FEYNMAN_DATA_CARD_ID:
            {
                card_id = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                break; 
            }
            case FEYNMAN_DATA_SESSION_ID:
            {
                session_id = (buf[i] << 8) | (buf[i+1] << 0);
                break; 
            }
            case FEYNMAN_DATA_SID:
            {
                sid = (buf[i] << 8) | (buf[i+1] << 0);
                break; 
            }
            case FEYNMAN_DATA_ECM:
            {
                memcpy(ecm_data, buf+i, nanolen);
                ecm_len = nanolen;
                break; 
            }
        }
        i += nanolen;
    }

    t_feynman_card *card = feynman_get_local_card(server, card_id);
    if(card !=NULL && ecm_len > 0 && sid > 0 && session_id > 0)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] New ECM from %s\n", client->username);
        t_feynman_ecm *ecm = feynman_add_local_ecm(server);
        ecm->caid = card->caid;
        ecm->provid = card->provid;
        ecm->sid = sid;
        ecm->session = session_id;
        ecm->decrypt_status = FEYNMAN_ECMSTATUS_WAITING;
        ecm->client = client;
        ecm->card_id = card_id;
        feynman_set_ecm_data(ecm, ecm_data, ecm_len);

        int ecm_status = server->ecm_callback(server->ecm_callback_args, ecm);
        if(ecm_status != FEYNMAN_ECMSTATUS_WAITING)
        {
            ecm->decrypt_status = ecm_status;
            feynman_send_ecm_ack(ecm);
        }
        return FEYNMAN_STATUS_OK;
    }
    else
    {
        char *reason = "Unknown reason";

        if(card == NULL) reason = "Card not avaliable";
        if(ecm_len <= 0) reason = "Invalid ECM length";
        if(sid <= 0) reason = "Invalid SID";
        if(session_id <= 0) reason = "Invalid session";

        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Bad ECM from %s (%s)\n", client->username, reason);
    }

    return FEYNMAN_ERROR;
}

int feynman_send_ecm_ack(t_feynman_ecm *ecm)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];
    if(ecm == NULL) return FEYNMAN_ERROR;

    t_feynman_client *client = ecm->client;
    if(client == NULL) return FEYNMAN_ERROR;

    t_feynman_local_server *server = client->server;
    if(server == NULL) return FEYNMAN_ERROR;

    int i = 0;

    send_buf[i++] = FEYNMAN_MSG_ECM_ACK;
    send_buf[i++] = 0;
    send_buf[i++] = 0;

    send_buf[i++] = FEYNMAN_DATA_CARD_ID;
    send_buf[i++] = 4;
    send_buf[i++] = (ecm->card_id >> 24) & 0xFF;
    send_buf[i++] = (ecm->card_id >> 16) & 0xFF;
    send_buf[i++] = (ecm->card_id >> 8) & 0xFF;
    send_buf[i++] = ecm->card_id & 0xFF;

    if(ecm->dcw != NULL && ecm->dcw_len > 0)
    {
        send_buf[i++] = FEYNMAN_DATA_DCW;
        send_buf[i++] = ecm->dcw_len;
        memcpy(send_buf+i, ecm->dcw, ecm->dcw_len);
        i += ecm->dcw_len;
    }

    send_buf[i++] = FEYNMAN_DATA_SESSION_ID;
    send_buf[i++] = 2;
    send_buf[i++] = (ecm->session >> 8) & 0xFF;
    send_buf[i++] = ecm->session & 0xFF;

    send_buf[1] = ((i-3) >> 8);
    send_buf[2] = ((i-3) >> 0);

    int session = ecm->session;
    int card_id = ecm->card_id;

    feynman_del_local_ecm(server, ecm);

    if(feynman_send_network_message(client->handle, send_buf, i) < 0) return FEYNMAN_ERROR;
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] ECM ack for session %d (card %X) sent to %s\n", session, card_id, client->username);
    return FEYNMAN_STATUS_OK;
}

int feynman_process_ecm_ack(t_feynman_remote_server *server, unsigned char *buf, int len)
{
    int i;
    int session_id = -1;
    int card_id = -1;
    unsigned char dcw[16];
    int dcw_len = 0;
    struct timeval time_diff;

    if(server == NULL) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];

        switch(nano)
        {
            case FEYNMAN_DATA_DCW:
            {
                memcpy(dcw, buf+i, nanolen);
                dcw_len = nanolen;
                break; 
            }
            case FEYNMAN_DATA_SESSION_ID:
            {
                session_id = (buf[i] << 8) | (buf[i+1] << 0);
                break;
            }
            case FEYNMAN_DATA_CARD_ID:
            {
                card_id = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                break; 
            }
        }
        i += nanolen;
    }

    if(session_id < 0 ) return FEYNMAN_ERROR;
    if(card_id < 0 )
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] No card id sent with dcw, old version of server?\n");
        return FEYNMAN_ERROR;
    }

    t_feynman_ecm *ecm = feynman_get_remote_ecm(server, session_id, card_id);
    if(ecm == NULL) return FEYNMAN_ERROR;

    gettimeofday(&time_diff, NULL);
    feynman_get_time_diff(&ecm->time_added, &time_diff);

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Got session %d (card %X) back from server (%ds %dus)\n", session_id, card_id, time_diff.tv_sec, time_diff.tv_usec);

    if(dcw_len > 0)
    {
        ecm->decrypt_status = FEYNMAN_ECMSTATUS_DECRYPT_OK;
        feynman_set_ecm_dcw(ecm, dcw, dcw_len);
    }
    else
    {
        ecm->decrypt_status = FEYNMAN_ECMSTATUS_DECRYPT_NOK;
    }

    if(server->dcw_callback != NULL)
    {
        server->dcw_callback(server->dcw_callback_args, ecm);
    }
    return FEYNMAN_STATUS_OK;
}

int feynman_send_emm(t_feynman_remote_server *server, t_feynman_emm *emm)
{
    unsigned char send_buf[FEYNMAN_BUF_SIZE];
    int i = 0;

    if(server == NULL) return FEYNMAN_ERROR;
    if(emm == NULL) return FEYNMAN_ERROR;

    send_buf[i++] = FEYNMAN_MSG_EMM;
    send_buf[i++] = 0;
    send_buf[i++] = 0;

    send_buf[i++] = FEYNMAN_DATA_CARD_ID;
    send_buf[i++] = 4;
    send_buf[i++] = (emm->card_id >> 24) & 0xFF;
    send_buf[i++] = (emm->card_id >> 16) & 0xFF;
    send_buf[i++] = (emm->card_id >> 8) & 0xFF;
    send_buf[i++] = emm->card_id & 0xFF;

    send_buf[i++] = FEYNMAN_DATA_EMM_TYPE;
    send_buf[i++] = 1;
    send_buf[i++] = emm->type & 0xFF;

    send_buf[i++] = FEYNMAN_DATA_EMM_ADDRESS;
    send_buf[i++] = 4;
    send_buf[i++] = (emm->address >> 24) & 0xFF;
    send_buf[i++] = (emm->address >> 16) & 0xFF;
    send_buf[i++] = (emm->address >> 8) & 0xFF;
    send_buf[i++] = emm->address & 0xFF;

    send_buf[i++] = FEYNMAN_DATA_EMM;
    send_buf[i++] = emm->emm_len;
    memcpy(send_buf+i, emm->emm, emm->emm_len);
    i += emm->emm_len;

    send_buf[1] = ((i-3) >> 8);
    send_buf[2] = ((i-3) >> 0);

    if(feynman_send_network_message(server->handle, send_buf, i) < 0) return FEYNMAN_ERROR;
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] EMM sent to card %X, caid: %04X provid:%08X address: %08X type: %d\n", emm->card_id, emm->caid, emm->provid, emm->address, emm->type);
    return FEYNMAN_STATUS_OK;
}

int feynman_process_emm(t_feynman_client *client, t_feynman_local_server *server, unsigned char *buf, int len)
{
    unsigned char emm_data[FEYNMAN_BUF_SIZE];

    int card_id = -1;
    int emm_len = -1;
    int address = 0;
    int address_ok = 0;
    int type = 0;
    int i;

    if(client == NULL) return FEYNMAN_ERROR;
    if(server == NULL) return FEYNMAN_ERROR;

    if(!(client->access_mode & FEYNMAN_ACCESS_MODE_EMM))
    {
        return FEYNMAN_STATUS_NOK;
    }

    if(server->emm_callback == NULL) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = buf[i++];

        switch(nano)
        {
            case FEYNMAN_DATA_CARD_ID:
            {
                card_id = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                break; 
            }
            case FEYNMAN_DATA_EMM:
            {
                memcpy(emm_data, buf+i, nanolen);
                emm_len = nanolen;
                break; 
            }
            case FEYNMAN_DATA_EMM_TYPE:
            {
                type = buf[i];
                break; 
            }
            case FEYNMAN_DATA_EMM_ADDRESS:
            {
                if((client->access_mode & type))
                {
                    address = (buf[i] << 24) | (buf[i+1] << 16) | (buf[i+2] << 8) | (buf[i+3] << 0);
                }
                break; 
            }
        }
        i += nanolen;
    }

    if(type == 0 || address == 0) return FEYNMAN_STATUS_NOK;

    t_feynman_card *card = feynman_get_local_card(server, card_id);

    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Got EMM type %d (%04X:%08X - %08X) for card %X\n", type, card->caid, card->provid, address, card->card_id);

    if(card !=NULL && emm_len > 0)
    {
        if((client->access_mode & FEYNMAN_ACCESS_MODE_UA) && type == FEYNMAN_ADDRESS_TYPE_UA)
        {
            if(address != card->ua)
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] UA missmatch\n");
                return FEYNMAN_ERROR;
            }
            else
            {
                address_ok = 1;
            }
        }
        else if((client->access_mode & FEYNMAN_ACCESS_MODE_SA) && type == FEYNMAN_ADDRESS_TYPE_SA)
        {
            if(address != card->sa)
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] SA missmatch\n");
                return FEYNMAN_ERROR;
            }
            else
            {
                address_ok = 1;
            }
        }
        else if((client->access_mode & FEYNMAN_ACCESS_MODE_GA) && type == FEYNMAN_ADDRESS_TYPE_GA)
        {
            if(address != card->ga)
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] GA missmatch\n");
                return FEYNMAN_ERROR;
            }
            else
            {
                address_ok = 1;
            }
        }

        if(!address_ok)
        {
            feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Invalid EMM address %X (type %d)\n", address, type);
            return FEYNMAN_ERROR;
        }

        t_feynman_emm *emm = feynman_add_local_emm(NULL);
        emm->caid = card->caid;
        emm->provid = card->provid;
        emm->address = address;
        emm->type = type;
        emm->card_id = card_id;
        feynman_set_emm_data(emm, emm_data, emm_len);

        server->emm_callback(server->emm_callback_args, emm);
        feynman_free_emm(emm);

        return FEYNMAN_STATUS_OK;
    }

    return FEYNMAN_ERROR;
}

int feynman_process_client_msg(t_feynman_local_server *server, t_feynman_handle *handle, unsigned char *buf, int len)
{
    int i;

    if(server == NULL) return FEYNMAN_ERROR;

    t_feynman_client *client = feynman_get_client(server, handle);

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = (buf[i++] << 8);
        nanolen |= buf[i++];

        switch(nano)
        {
            case FEYNMAN_MSG_LOGIN:
            {
                client = feynman_process_login(server, buf+i, nanolen);
                if(feynman_send_login_ack(client, handle) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while sending FEYNMAN_MSG_LOGIN_ACK\n");
                    break; 
                }

                if(client == NULL) return FEYNMAN_STATUS_NOK;

                t_feynman_card *card;
                for(card=server->card_db; card!=NULL; card=card->next)
                {
                    if(feynman_send_add_card(client, card) == FEYNMAN_ERROR)
                    {
                        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while sending FEYNMAN_MSG_CARD\n");
                        break; 
                    }
                }
                for(card=server->card_db; card!=NULL; card=card->next)
                {
                    if(feynman_send_sidlist(client, card) == FEYNMAN_ERROR)
                    {
                        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while sending FEYNMAN_MSG_SID\n");
                        break; 
                    }
                }
                break; 
            }
            case FEYNMAN_MSG_ECM:
            {
                if(feynman_process_ecm(client, server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while sending FEYNMAN_MSG_ECM_ACK\n");
                    break; 
                }
                break; 
            }
            case FEYNMAN_MSG_EMM:
            {
                if(feynman_process_emm(client, server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while processing FEYNMAN_MSG_EMM\n");
                    break; 
                }
                break; 
            }
            default:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Unknown nano %X\n", nano);
                break;
            }
        }
        i += nanolen;
    }
    return FEYNMAN_STATUS_OK;
}

int feynman_process_remote_server_msg(t_feynman_remote_server *server, unsigned char *buf, int len)
{
    int i;
    if(server == NULL) return FEYNMAN_ERROR;

    for(i=0; i<len;)
    {
        int nano = buf[i++];
        int nanolen = (buf[i++] << 8);
        nanolen |= buf[i++];

        switch(nano)
        {
            case FEYNMAN_MSG_LOGIN_ACK:
            {
                if(feynman_process_login_ack(server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while processing FEYNMAN_MSG_LOGIN_ACK\n");
                }
                break;
            }
            case FEYNMAN_MSG_ADD_CARD:
            {
                if(feynman_process_add_card(server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while processing FEYNMAN_MSG_ADD_CARD\n");
                }
                break;
            }
            case FEYNMAN_MSG_DEL_CARD:
            {
                if(feynman_process_del_card(server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while processing FEYNMAN_MSG_DEL_CARD\n");
                }
                break;
            }
            case FEYNMAN_MSG_SIDS:
            {
                if(feynman_process_sids(server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while processing FEYNMAN_MSG_SIDS\n");
                }
                break;
            }
            case FEYNMAN_MSG_ECM_ACK:
            {
                if(feynman_process_ecm_ack(server, buf+i, nanolen) == FEYNMAN_ERROR)
                {
                    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error while processing FEYNMAN_MSG_ECM_ACK\n");
                }
                break;
            }
            default:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Unknown nano %X\n", nano);
                break;
            }

        }
        i += nanolen;
    }
    return FEYNMAN_STATUS_OK;
}

int feynman_send_server_ecm(t_feynman_remote_server *server, unsigned short session, int card_id, unsigned short caid, int provid, unsigned short sid, unsigned char *ecm_data, int len)
{
    int card_db[255];
    t_feynman_ecm *ecm = NULL;
    int ret = 0;

    if(server == NULL) return ret;
    if(!server->ready) return ret;

    srand((unsigned int)time(0));
    feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Trying to use old card id %X for %04X:%08X\n", card_id, caid, provid);
    feynman_lock_remote_server(server);

    t_feynman_card *card = feynman_get_remote_card(server, card_id);
    if(card == NULL)
    {
        card_id = -1;
    }
    else
    {
        card_id = card->card_id;

        if(card->caid != caid || card->provid != provid)
        {
          card_id = -1;
        }
        else
        {
            t_feynman_sid *tmp_sid = feynman_get_sid(card, sid);

            if(tmp_sid != NULL)
            {
                if(tmp_sid->mode == FEYNMAN_SIDMODE_BLOCK) card_id = -1;
            }
        }
    }

    if(card_id < 0)
    {
        card = NULL;
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] old card id is not good\n");
        int card_count = feynman_get_supprted_cards(server, card_db, caid, provid, sid);
        int i = 0;

        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Sending ecm to %d cards \n", card_count);

        for(i=0; i<card_count; i++)
        {
            ecm = feynman_add_remote_ecm(server);
            ecm->caid = caid;
            ecm->provid = provid;
            ecm->sid = sid;
            ecm->session = session;
            ecm->decrypt_status = FEYNMAN_ECMSTATUS_WAITING;
            ecm->card_id = card_db[i];
            feynman_set_ecm_data(ecm, ecm_data, len);
            feynman_send_ecm(server, ecm);
        }
        ret = 0;
    }

    if(card != NULL)
    {
        ecm = feynman_add_remote_ecm(server);
        ecm->caid = caid;
        ecm->provid = provid;
        ecm->sid = sid;
        ecm->session = session;
        ecm->decrypt_status = FEYNMAN_ECMSTATUS_WAITING;
        ecm->card_id = card_id++;
        feynman_set_ecm_data(ecm, ecm_data, len);
        feynman_send_ecm(server, ecm);
        ret = 1;
    }

    feynman_unlock_remote_server(server);
    return ret;
}

//Use only without dcw_callback!
int feynman_get_server_dcw(t_feynman_remote_server *server, unsigned short session, int *card_id, unsigned char *dcw)
{
    int ret = FEYNMAN_ECMSTATUS_POOL_EMPTY;
    t_feynman_ecm *ecm = NULL;

    if(server == NULL) return FEYNMAN_ECMSTATUS_ERROR;
    if(server->dcw_callback !=NULL) return FEYNMAN_ECMSTATUS_ERROR;

    feynman_lock_remote_server(server);

    for(ecm=server->ecm_db; ecm!=NULL; ecm=ecm->next)
    {
        if(ecm->session == session)
        {
            ret = ecm->decrypt_status;
            if(ret == FEYNMAN_ECMSTATUS_DECRYPT_OK)
            {
                memcpy(dcw, ecm->dcw, ecm->dcw_len);
            }
            if(ret != FEYNMAN_ECMSTATUS_WAITING)
            {
                *card_id = ecm->card_id;
                feynman_del_remote_ecm(server, ecm);
                break;
            }
        }
    }

    feynman_unlock_remote_server(server);
    return ret;
}

int feynman_dcw_dummy_callback(t_feynman_remote_server *server, unsigned short session, unsigned char *dcw, int timeout_ms)
{
    int card_id = -1;
    if(server == NULL) return card_id;

    int timeout = 0;
    while(1)
    {
        if(timeout > timeout_ms)
        {
            feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Timeout waiting for session %d\n", session);
            return -1;
            break;
        }
        int decrypt_status = feynman_get_server_dcw(server, session, &card_id, dcw);
        switch(decrypt_status)
        {
            case FEYNMAN_ECMSTATUS_WAITING:
            {
                break;
            }
            case FEYNMAN_ECMSTATUS_DECRYPT_OK:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Decrypt OK on session %d (card %X)\n", session, card_id);
                return card_id;
            }
            case FEYNMAN_ECMSTATUS_DECRYPT_NOK:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Decrypt failed on session %d (card %X)\n", session, card_id);
                break; 
            }
            case FEYNMAN_ECMSTATUS_ERROR:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Error on session %d\n", session);
                return -1;
            }
            case FEYNMAN_ECMSTATUS_IDLE:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] Ecm on session %d is idle, shouldn't happend\n", session);
                return -1;
            }
            case FEYNMAN_ECMSTATUS_POOL_EMPTY:
            {
                feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] No cards could decrypt the ecm on this server\n");
                return -1;
            }
        }
        usleep(10000);
        timeout += 10;
    }
}

//TODO: set sessions somehow
void feynman_send_server_emm(t_feynman_remote_server *server, unsigned short caid, int provid, int address, unsigned char *emm_data, int len)
{
    t_feynman_card *card;
    t_feynman_emm *emm = NULL;
    int emm_sent = 0;
    if(!server) return;
    if(!server->ready) return;
    if(!address) return;

    feynman_lock_remote_server(server);

    for(card=server->card_db; card!=NULL; card=card->next)
    {
        if(card->caid == caid && card->provid == provid)
        {
            if(card->ga == address || card->sa == address || card->ua == address)
            {
                emm = feynman_add_remote_emm(server);
                emm->caid = caid;
                emm->provid = provid;
                emm->address = address;
                if(card->ga == address) emm->type = FEYNMAN_ADDRESS_TYPE_GA;
                else if(card->sa == address) emm->type = FEYNMAN_ADDRESS_TYPE_SA;
                else if(card->ua == address) emm->type = FEYNMAN_ADDRESS_TYPE_UA;
                emm->card_id = card->card_id;
                feynman_set_emm_data(emm, emm_data, len);
                feynman_send_emm(server, emm);
                emm_sent = 1;
            }
        }
    }

    if(!emm_sent)
    {
        feynman_logging(FEYNMAN_LOGMODE_ALL, "[ Feynman ] No card found for address %X\n", address);
    }

    feynman_unlock_remote_server(server);
}
